%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>

#include <sys/ioctl.h>
#if !defined (FIONREAD)
#	include <sys/filio.h>		/* LINUX does not have filio.h */
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

#include <kqml/kqmlparse.h>

static void syntax_error (char *s, char *token);
static void semantic_error (char *s, char *token);
static void memory_error (char *s, char *token, int len);
static void incomplete_message (char *s);

static void parse_init (void);
static void parse_started (void);
static int parse_error (void);
static int parse_exit (void);

static int set_parameter (char *s);
static char *get_parameter ();
static void clear_parameter ();

static void set_value_type (kqml_expression_type value_type);
static kqml_expression_type get_value_type (void);
static void clear_value_type (void);

static void add_performative (char *s);
static void add_parameter (char *s);
static void add_value (void);

static int check_eof (void);
static int do_parse (kqml_message **msg);
%}

%token START_TOKEN STRING BYTE_STRING WORD KEYWORD INVALID_SYM
%token '(' ')' ',' '`' '\''

%%

incoming_message:
	performative
		{
			return parse_exit ();
		}
	| /* null */
		{
  			return parse_error ();
		}
	| error
		{
			return parse_error ();
		}
;

performative :
        {
			  parse_init ();
        }
	optional_start_token '(' 
			{
				parse_started ();
			}
		WORD
        { 
				add_performative (yylval);
        }
      keyword_list ')'
        {
				return parse_exit ();
        }
   | error
        { 
			return parse_error ();
        }
;

optional_start_token :
     /* empty */
   | START_TOKEN
;

keyword_list :
     /* empty */
   | keyword_list
     WORD
        { 
			 add_parameter (yylval);
			 if (yylval [0] != ':')
			 {
				semantic_error ("No ':' in keyword", yylval);
			 }
        }
     expression
			{
          add_value ();
        	}
;

expression :
     WORD
		{
			set_value_type (KQML_TYPE_WORD);
		}
   | STRING
		{
			set_value_type (KQML_TYPE_STRING);
		}
	| BYTE_STRING
		{
			set_value_type (KQML_TYPE_BYTE);
		}
   | quotation
   | s_expression
;

s_expression :
     '(' 
			{
				set_value_type (KQML_TYPE_SEXPR);
			}
	  expression_list ')'
;

expression_list :
     /* empty */
   | expression_list expression_list_term
;

expression_list_term :
     expression
   | comma_expression

quotation :
     '\''
		{
			set_value_type (KQML_TYPE_QUOTED);
		}
		expression
   | '`'
		{
			set_value_type (KQML_TYPE_COMMA);
		}
		expression
   | '`'
		{
			set_value_type (KQML_TYPE_COMMA);
		}
		comma_expression
;

comma_expression :
     ',' WORD
   | ',' s_expression
;

%%

void
yyerror (s)
   char *s;
{
   if (yychar == START_TOKEN)
   {
      /* if the start token is detected anywhere in a kqml sentence other
         than at the very beginning of a message, it is a sign that another
         sentence is beginning without finding the end of the previous one */
		incomplete_message ("Incomplete message");
   }
   else
   {
		syntax_error ("Syntax error scanning", yylval);
   }
}

typedef enum { 
	KQML_PARSE_FROM_YYIN = 0, KQML_PARSE_FROM_FD, KQML_PARSE_FROM_STREAM,
	KQML_PARSE_FROM_BUFFER 
	} parse_input_type;

static int syntax_error_count = 0, semantic_error_count = 0,
	memory_error_count = 0, incomplete_message_count = 0;

static kqml_expression_type value_type = KQML_TYPE_UNKNOWN;
static kqml_message *kmsg = NULL;
static char *parameter = NULL;

static parse_input_type input_from = KQML_PARSE_FROM_YYIN;
static int input_count = 0;
static void *input_var = NULL;
static char input_started = FALSE, parser_timed_out = FALSE;

static void
syntax_error (char *s, char *token)
{
	syntax_error_count++;
	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_parse: %s \"%s\".", s, token);
}

static void
semantic_error (char *s, char *token)
{
   semantic_error_count++;
   _kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_parse: %s \"%s\".", s, token);
}

static void
memory_error (char *s, char *token, int len)
{
	memory_error_count++;
	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_parse: %s \"%.*s\".", s, len, token);
}

static void
incomplete_message (char *s)
{
	incomplete_message_count++;
	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_parse: %s.", s);
}

static void
parse_init ()
{

#	ifdef YYDEBUG
		yydebug = 1;
#	endif

	syntax_error_count =
		semantic_error_count =
		memory_error_count =
		incomplete_message_count = 0;
	kmsg = NULL;
	parameter = NULL;
}

static void
parse_started ()
{
	input_started = TRUE;
}

static int
parse_error ()
{
	int ret_value;

	if (check_eof ())
	{
		/* If we receive eof, it is definitely timeout,
			a genuine eof should have been discovered by 
			the check_eof call in do_parse */
		if (parser_timed_out)
		{
			ret_value = KQML_PARSE_TIMEOUT;
		}
		else if (! input_started)
		{
			ret_value = KQML_PARSE_NOTHING;
		}
		else 
		{
			ret_value = KQML_PARSE_EOF;
		}
	}
	else if (incomplete_message_count)
	{
		ret_value = KQML_PARSE_INCOMPLETE;
	}
	else if (syntax_error_count)
	{
		ret_value = KQML_PARSE_SYNTAX;
	}
	else if (semantic_error_count)
	{
		ret_value = KQML_PARSE_SEMANTIC;
	}
	else if (memory_error_count)
	{
		ret_value = KQML_PARSE_MEMORY;
	}
	else
	{
		/* Should not occur. But possible (remotely) if we timeout
			and then data comes in */
		ret_value = KQML_PARSE_TIMEOUT;
	}

	input_from = KQML_PARSE_FROM_YYIN;
	input_count = 0;
	input_started = FALSE;
	kqml_free_msg (kmsg);
	kmsg = NULL;
	clear_parameter ();
	return ret_value;
}

static int
parse_exit ()
{
	input_from = KQML_PARSE_FROM_YYIN;
	input_count = 0;
	input_started = FALSE;
	return KQML_PARSE_OK;
}

static int
set_parameter (char *s)
{
	return ((parameter = strdup (s)) != NULL);
}

static char *
get_parameter ()
{
	return parameter;
}

static void
clear_parameter ()
{
	if (parameter)
	{
		free (parameter);
		parameter = NULL;
	}
}

static void
set_value_type (kqml_expression_type new_type)
{
	if (value_type == KQML_TYPE_UNKNOWN)
	{
		value_type = new_type;
	}
}

static kqml_expression_type
get_value_type ()
{
	return value_type;
}

static void
clear_value_type ()
{
	value_type = KQML_TYPE_UNKNOWN;
}


static void
add_performative (char *s)
{
	clear_parameter ();
	kmsg = kqml_build_msgv (s, NULL, NULL, NULL, NULL, NULL);
   if (kmsg == NULL)
	{
		memory_error ("Not enough memory for kmsg", "", 0);
		return;
	}

	return;
}

static void
add_parameter (char *s)
{
	_kqml_parse_start_recording_value ();
	clear_value_type ();

	if (kmsg)
	{
		if (! set_parameter (s))
		{
			kqml_free_msg (kmsg);
			kmsg = NULL;
			memory_error ("Not enough memory for parameter", s, strlen(s));
			return ;
		}
	}

	return;
}

static void
add_value ()
{
	char *temp, *value, *field; 
	int value_len;
	kqml_expression_type value_type;

	_kqml_parse_stop_recording_value (&temp, &value_len);
	if (kmsg)
	{
		value_type = get_value_type ();
		field = get_parameter ();

		if (temp == NULL)
		{
			memory_error ("Not enough memory for internal buffer", "", 0);
			kqml_free_msg (kmsg);
			kmsg = NULL;
			clear_parameter ();
			return;
		}

		if (value_type == KQML_TYPE_STRING)
		{
			temp++;
			value_len -= 2;
		}
		else if (value_type == KQML_TYPE_BYTE)
		{
			char *dummy;

			dummy = memchr(temp, (int)'"', value_len);
			value_len -= (dummy - temp) + 1;
			temp = dummy + 1;
		}

		value = malloc (value_len + 1);
		if (value == NULL)
		{
			kqml_free_msg (kmsg);
			kmsg = NULL;
			clear_parameter ();
			memory_error ("Not enough memory for value", temp, value_len);
			return;
		}

		memcpy (value, temp, value_len);
		value[value_len] = '\0';

		if (_kqml_put_field_nd (field, kmsg, value, value_len, value_type) <= -1)
		{
			kqml_free_msg (kmsg);
			free (value);
			clear_parameter ();
			memory_error ("Not enough memory for param value pair", temp, 
				value_len);
			return;
		}
	}

	clear_parameter ();
	return;
}

static int
do_parse (kqml_message **msg)
{
	int ret_value;

	kmsg = NULL;
	input_started = FALSE;
	parser_timed_out = FALSE;

	if (check_eof ())
	{
		input_from = KQML_PARSE_FROM_YYIN;
		*msg = NULL;
      return KQML_PARSE_EOF;
	}
	else
	{
   	ret_value = yyparse ();
		*msg = kmsg;
		kmsg = NULL;
		return ret_value;
	}
}

int kqml_parse_fd (int fd, kqml_message **msg)
{
	input_from = KQML_PARSE_FROM_FD;
	(int *) input_var = &fd;
	input_count = 0;

	return do_parse (msg);
}

int kqml_parse_stream (FILE *fp, kqml_message **msg)
{
	input_from = KQML_PARSE_FROM_STREAM;
	(FILE *) input_var = fp;
	input_count = 0;

	return do_parse (msg);
}

int kqml_parse_buffer (char *buffer, int len, int *used_len, kqml_message **msg)
{
	int ret_value;

	input_from = KQML_PARSE_FROM_BUFFER;
	(char *) input_var = buffer;
	if (buffer == NULL)
	{
		input_count = 0;
	}
	else
	{
		len = (len < 0 ? strlen (buffer) : len);
		input_count = len;
	}

	ret_value = do_parse (msg);
	if (used_len)
	{
		*used_len = len - input_count;
	}
	return ret_value;
}

static int
check_eof ()
{
	switch (input_from)
	{
	case KQML_PARSE_FROM_FD:
	{
		int fd;

		if (input_var == NULL)
			return TRUE;

		fd = *((int *)input_var);
		return (input_count = _kqml_pending_bytes (fd, NULL)) == 0;

		break;
	}
	case KQML_PARSE_FROM_STREAM:
	{
		FILE *fp;

		if (input_var == NULL)
			return TRUE;

		fp = (FILE *)input_var;

		return (input_count = _kqml_pending_bytes (fileno (fp), fp)) == 0;

		break;
	}
	case KQML_PARSE_FROM_BUFFER:
	{
		if (input_var == NULL)
			return TRUE;

		return input_count == 0;
		
		break;
	}
	default:
	{
		return feof (stdin);

		break;
	}
	}
}

int
_kqml_parse_get_char ()
{
	int i;

	switch (input_from)
	{
	case KQML_PARSE_FROM_FD:
	{
		int fd;
		char c;

		if (input_count == 0)
		{
			if (input_started)
			{
				for (i = 0; i < KQML_PARSER_WAIT_TIME; i++)
				{
					if (! check_eof ())
						break;
					sleep(1);
				}
				if (i == KQML_PARSER_WAIT_TIME)
				{
					parser_timed_out = TRUE;
					return EOF;
				}
			}
			else
				return EOF;
		}

		fd = *((int *)input_var);
		
		if (read (fd, &c, 1) == 0)
		{
			input_count = 0;
			return EOF;
		}
		else
		{
			input_count--;
			return c;
		}

		break;
	}
	case KQML_PARSE_FROM_STREAM:
	{
		FILE *fp;
		int c;

		if (input_count == 0)
		{
			if (input_started)
			{
				for (i = 0; i < KQML_PARSER_WAIT_TIME; i++)
				{
					if (! check_eof ())
						break;
					sleep (1);
				}
				if (i == KQML_PARSER_WAIT_TIME)
				{
					parser_timed_out = TRUE;
					return EOF;
				}
			}
			else
				return EOF;
		}

		fp = (FILE *)input_var;

		c = getc (fp);
		if (c == EOF)
		{
			input_count = 0;
			return EOF;
		}
		else
		{
			input_count--;
			return c;
		}

		break;
	}
	case KQML_PARSE_FROM_BUFFER:
	{
		char *buffer;
		int c;

		buffer = (char *)input_var;

		if (input_count == 0 || buffer == NULL)
			return EOF;

		c = *buffer;
		if (--input_count == 0)
		{
			(char *)input_var = NULL;
		}
		else
		{
			(char *)input_var = ++buffer;
		}
		return c;
	
		break;
	}
	default:
	{
		return getc (stdin);

		break;
	}
	}
}
