#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <malloc.h>
#include <ctype.h>

#include <kqml.h>

#include <kqml/kqmlparse.h>

#include <kqml/kqml.tab.h>

/* 
	NOTE: If START_TOKEN_STR is modified,
	the logic of read_word should be examined thoroughly
	Do update the START_TOKEN_LEN please.
*/
#define START_TOKEN_LEN 5
#define START_TOKEN_STR "!@#$%"

static void read_whitespace (int c, int *next_char, int *len);
static int read_word (int c, int *next_char, int *len);
static int read_bounded_token (int c, int *next_char, int *len);
static int read_string_token (int c, int *next_char, int *len);

static int first_value ();
static void add_token (int len, char *ptr);
static void set_token_ptr ();
static char *get_token_ptr (int len);

extern char *yylval;

int yyleng;

int
yylex()
{
	static int next_char = EOF;
	int c;
	int ret_val;

	yylval = NULL;
	do
	{
		if (next_char == EOF)
		{
			c = _kqml_parse_get_char ();
		}
		else
		{
			c = next_char;
			next_char = EOF;
		}

		if (c == EOF)
		{
			return EOF;
		}

		if (isspace(c))
		{
			read_whitespace (c, &next_char, &yyleng);
		}
		else if (c == '(' || c == ')' || c == ',' || c == '`' || c == '\'')
		{
			char temp[1];

			set_token_ptr ();
			yyleng = 1;
			temp[0] = c;
			add_token (1, temp);
			yylval = get_token_ptr (yyleng);
			return c;
		}
		else if (c == '"')
		{
			ret_val = read_string_token (c, &next_char, &yyleng);
			yylval = get_token_ptr (yyleng);
			return ret_val;
		}
		else if (c == '#')
		{
			ret_val = read_bounded_token (c, &next_char, &yyleng);
			yylval = get_token_ptr (yyleng);
			return ret_val;
		}
		else 
		{
			/* It can be a word or a start_token or invalid */
			ret_val = read_word (c, &next_char, &yyleng);
			yylval = get_token_ptr (yyleng);
			return ret_val;
		}
	}
	while (TRUE);
}

yywrap ()
{
   return (1);
}

static void
read_whitespace (int c, int *next_char, int *len)
{
	*len = 0;

	/* do not add leading whitespace to the value buffer */
	if (! first_value ())
	{
		add_token (1, " ");
	}

	do
	{
		c = _kqml_parse_get_char ();
		if (c == EOF)
		{
			break;
		}
	}
	while (isspace (c));

	*next_char = c;
}

static int
read_word (int c, int *next_char, int *len)
{
	static char word_table[256], initialised = FALSE;
	int i, token_len, buffer_len;
	char token_buffer[256];
	char valid_word = TRUE, end_of_word = FALSE, valid_token = FALSE;

	if (! initialised)
	{
		char *str1, *str2;

		str1 = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
		str2 = "<>=+-*/&^~_@$%:.!?";
		
		memset(word_table, 0, sizeof(word_table));
		while (*str1)
		{
			word_table[(int)*str1] = TRUE;
			str1++;
		}
		while (*str2)
		{
			word_table[(int)*str2] = TRUE;
			str2++;
		}
		initialised = TRUE;
	}

	/* We get the address of the pointer which points to the memory block
		'cos during intermediate realloc we might be allocated a totally
		new block. So, we should get the actual address only when we have
		completed 
	*/

	set_token_ptr ();
	*len = 0;
	token_len = buffer_len = 0;
	token_buffer[buffer_len++] = c;

	if (c == START_TOKEN_STR[0])
	{
		for (i = 1; i < START_TOKEN_LEN; i++)
		{
			if (buffer_len == 256)
			{
				add_token (buffer_len, token_buffer);
				token_len += buffer_len;
				buffer_len = 0;
			}

			c = _kqml_parse_get_char ();
			if (c == EOF)
			{
				end_of_word = TRUE;
				break;
			}

			if (c != START_TOKEN_STR[i])
			{
				/* We have found out that this is not a start_token.
					If valid_word is TRUE, we can continue reading provided
						the current character also belongs to the word character
						set.
					Otherwise we have to exit.
				*/
				if (valid_word)
				{
					if (! word_table[c])
					{
						*next_char = c;
						end_of_word = TRUE;
					}
					else
					{
						token_buffer[buffer_len++] = c;
					}
				}
				break;
			}
			else
			{
				/* Set the valid_word flag */
				valid_word = (word_table[c] ? valid_word : FALSE);
				token_buffer[buffer_len++] = c;
			}
		}

		if (i == START_TOKEN_LEN)
		{
			end_of_word = TRUE;
			valid_token = TRUE;
		}
	}
	else
	{
		valid_word = word_table [c];
	}

	if (valid_word && ! end_of_word)
	{
		while (TRUE)
		{
			if (buffer_len == 256)
			{
				add_token (buffer_len, token_buffer);
				token_len += buffer_len;
				buffer_len = 0;
			}

			c = _kqml_parse_get_char ();
			if (c == EOF)
			{
				break;
			}

			if (word_table[c])
			{
				token_buffer[buffer_len++] = c;
			}
			else
			{
				*next_char = c;
				break;
			}
		}
	}

	if (buffer_len != 0)
	{
		add_token (buffer_len, token_buffer);
		token_len += buffer_len;
		buffer_len = 0;
	}
	*len = token_len;
	if (valid_token)
	{
		return START_TOKEN;
	}
	else
	{
		return (valid_word ? WORD : INVALID_SYM);
	}
}

static int 
read_bounded_token (int c, int *next_char, int *len)
{
   int token_len, buffer_len, string_len, num_len;
	char token_buffer[256], valid_string;

	set_token_ptr ();
	*len = 0;
	token_len = buffer_len = 0;
	token_buffer[buffer_len++] = '#';
	valid_string = TRUE;

	/* Read the no */
	string_len = 0;
	num_len = 0;
	do
	{
		if (buffer_len == 256)
		{
			add_token (buffer_len, token_buffer);
			token_len += buffer_len;
			buffer_len = 0;
		}

		c = _kqml_parse_get_char ();

		/* We have to save it immaterial of whether it is a digit or not.
			By doing so, we can find out what caused the error (yyerror 
			in kqml.y)
		*/
		if (c != EOF)
		{
			token_buffer[buffer_len++] = c;
		}
		if (c == '"')
		{
			/* If we have read no numbers, it is an error */
			valid_string = (num_len == 0 ? FALSE : TRUE);
			break;
		}
		if (c == EOF || ! isdigit (c))
		{
			valid_string = FALSE;
			string_len = 0; 
			break;
		}
		num_len++;
		string_len = string_len * 10 + c - '0';
	}
	while (TRUE);

	while (string_len--)
	{
		if (buffer_len == 256)
		{
			add_token (buffer_len, token_buffer);
			token_len += buffer_len;
			buffer_len = 0;
		}
		c = _kqml_parse_get_char ();
		if (c == EOF)
		{
			valid_string = FALSE;
			break;
		}
		token_buffer[buffer_len++] = c;
	}

	if (buffer_len != 0)
	{
		add_token (buffer_len, token_buffer);
		token_len += buffer_len;
		buffer_len = 0;
	}
	*len = token_len;
	return (valid_string ? BYTE_STRING : INVALID_SYM);
}

static int 
read_string_token (int c, int *next_char, int *len)
{
   int c2, buffer_len, token_len;
	char token_buffer[256], valid_string;

	set_token_ptr ();
	*len = 0;
	token_len = buffer_len = 0;
	token_buffer[buffer_len++] = '"';
	valid_string = TRUE;

	do
	{
		/* We may read 2 characters now. So, if we do not have space for
			2 characters, flush it 
		*/
		if (buffer_len >= 255)
		{
			add_token (buffer_len, token_buffer);
			token_len += buffer_len;
			buffer_len = 0;
		}

		c = _kqml_parse_get_char ();
		if (c == EOF)
		{
			valid_string = FALSE;
			break;
		}
		token_buffer[buffer_len++] = c;

		if (c == '\\')
		{
			c2 = _kqml_parse_get_char ();
			if (c2 == EOF)
			{
				valid_string = FALSE;
				break;
			} 
			token_buffer[buffer_len++] = c2;
		}
	}
	while (c != '"');

	if (buffer_len != 0)
	{
		add_token (buffer_len, token_buffer);
		token_len += buffer_len;
		buffer_len = 0;
	}
	*len = token_len;
	return (valid_string ? STRING : INVALID_SYM);
}

static int recording_value = FALSE;
static int value_len = 0;
static int max_value_len = 0;
static char *value = NULL;

static int
first_value ()
{
	return value_len == 0;
}

void
_kqml_parse_start_recording_value ()
{
   /* start buffering the value strings assuming that the temp_value_buffer
      will be big enough to contain the entire value */
	value_len = 0;
	recording_value = TRUE;
}

void
_kqml_parse_stop_recording_value (char **s, int *len)
{
	if (value_len > 0)
	{
		*s = value;
		*len = value_len;
	}
	else
	{
		*s = NULL;
		*len = 0;
	}
	recording_value = FALSE;
	value_len = 0;
}

/*
	This function should be once before each token is read. Ofcourse,
	after reading the first character of the token. Since, we are
	using the first character to decide which function to call.
	The exception is read_whitespace. It will not call this function.
	If we call this function, that means we are starting to process
	a new token. Based on what we are doing we should continue adding
	to the existing token in the buffer or throw the contents of the
	buffer and start all over again.

	white space is not a token, it is a separator. so, we do not call
	this function from read_whitespace.

	The other token reading functions call this once. 
	Based on recording_value, we either throw away the token in the
	buffer or continue adding to that token.
*/
static void
set_token_ptr ()
{
	/* We will overwrite the existing contents of the buffer, if
		we are not recording value i.e we are parsing the performative or
		the keyword. 
	*/
	if (! recording_value)
	{
		value_len = 0;
	}

	if (value == NULL)
	{
		max_value_len = 4096;
		value = malloc (max_value_len);
		
		if (value == NULL)
		{
			max_value_len = 0;
		}
	}
}

static char *
get_token_ptr (int token_len)
{
	if (value)
	{
		return (value + value_len - token_len);
	}
	else
	{
		return NULL;
	}
}

/* This function stores the token in our buffer, if necessary also
	extends the buffer 
*/

static void
add_token (int len, char *ptr)
{
	/* If we had encountered memory error for this value or if we do not
		have the value buffer or if we do not have to store anything
		return
	*/
	if (value_len < 0 || value == NULL || len <= 0)
	{
		return ;
	}

	/* The extra 1 is for the null character */
   if (value_len + len + 1 > max_value_len)
	{
		char *temp;
		int new_max_value_len;

		new_max_value_len = ((value_len + len + 1 + 511) / 512) * 512;

		temp = realloc (value, new_max_value_len);
		if (temp == NULL)
		{
			value_len = -1;
			return ;
		}

		value = temp;
		max_value_len = new_max_value_len;
	}

	memcpy (value+value_len, ptr, len);
	value_len += len;
	value[value_len] = '\0';
}
